apifishapifishapifish# -*- coding: utf-8 -*-
# Author: Arthur Imbert <arthur.imbert.pro@gmail.com>
# License: BSD 3 clause

"""
Unitary tests for apifish.detection.spot_modeling module.
"""


# TODO test apifish.detection.modelize_spot
# TODO test apifish.detection.initialize_grid
# TODO test apifish.detection.gaussian_2d
# TODO test apifish.detection.gaussian_3d
# TODO test apifish.detection.precompute_erf
# TODO test apifish.detection.fit_subpixel
