# -*- coding: utf-8 -*-
# Author: Arthur Imbert <arthur.imbert.pro@gmail.com>
# License: BSD 3 clause

"""
Unitary tests for apifish.classification.features module.
"""

# TODO add test apifish.classification.compute_features
# TODO add test apifish.classification.get_features_name
# TODO add test apifish.classification.features_distance
# TODO add test apifish.classification.features_in_out_nucleus
# TODO add test apifish.classification.features_protrusion
# TODO add test apifish.classification.features_dispersion
# TODO add test apifish.classification.features_topography
# TODO add test apifish.classification.features_foci
# TODO add test apifish.classification.features_area
# TODO add test apifish.classification.features_centrosome
