# -*- coding: utf-8 -*-
# Author: Arthur Imbert <arthur.imbert.pro@gmail.com>
# License: BSD 3 clause

"""
Unitary tests for apifish.classification.input_preparation module.
"""

# TODO add test apifish.classification.prepare_extracted_data
