apifishapifishapifish# -*- coding: utf-8 -*-
# Author: Arthur Imbert <arthur.imbert.pro@gmail.com>
# License: BSD 3 clause

"""
Unitary tests for apifish.multistack.colocalization module.
"""

# TODO add test apifish.multistack.detect_spots_colocalization
# TODO add test apifish.multistack.get_elbow_value_colocalized
