# -*- coding: utf-8 -*-
# Author: Arthur Imbert <arthur.imbert.pro@gmail.com>
# License: BSD 3 clause

"""
Unitary tests for apifish.segmentation.utils module.
"""


# TODO add test for apifish.segmentation.thresholding
# TODO add test for apifish.segmentation.compute_mean_diameter
# TODO add test for apifish.segmentation.compute_mean_convexity_ratio
# TODO add test for apifish.segmentation.compute_surface_ratio
# TODO add test for apifish.segmentation.count_instances
# TODO add test for apifish.segmentation.resize_image
# TODO add test for apifish.segmentation.get_marge_padding
# TODO add test for apifish.segmentation.compute_image_standardization
