apifishapifishapifish# -*- coding: utf-8 -*-
# Author: Arthur Imbert <arthur.imbert.pro@gmail.com>
# License: BSD 3 clause

"""
Unitary tests for apifish.detection.utils module.
"""

# TODO test apifish.detection.convert_spot_coordinates
# TODO add test for apifish.detection.get_object_radius_pixel
# TODO add test for apifish.detection.get_object_radius_nm
# TODO add test for apifish.detection.build_reference_spot
# TODO add test for apifish.detection.compute_snr_spots
# TODO add test for apifish.detection.get_breaking_point
