# -*- coding: utf-8 -*-
# Author: Arthur Imbert <arthur.imbert.pro@gmail.com>
# License: BSD 3 clause

"""
Unitary tests for apifish.segmentation.nuc_segmentation module.
"""


# TODO add test for apifish.segmentation.remove_segmented_nuc
# TODO add test for apifish.segmentation.unet_3_classes_nuc
# TODO add test for apifish.segmentation.apply_unet_3_classes
# TODO add test for apifish.segmentation.from_3_classes_to_instances
