# -*- coding: utf-8 -*-
# Author: Arthur Imbert <arthur.imbert.pro@gmail.com>
# License: BSD 3 clause

"""
Unitary tests for apifish.segmentation.postprocess module.
"""


# TODO add test for apifish.segmentation.label_instances
# TODO add test for apifish.segmentation.merge_labels
# TODO add test for apifish.segmentation.clean_segmentation
# TODO add test for apifish.segmentation.remove_disjoint
# TODO add test for apifish.segmentation.center_mask_coord
# TODO add test for apifish.segmentation.from_boundaries_to_surface
# TODO add test for apifish.segmentation.from_surface_to_boundaries
# TODO add test for apifish.segmentation.from_binary_to_coord
# TODO add test for apifish.segmentation.complete_coord_boundaries
# TODO add test for apifish.segmentation.from_coord_to_frame
# TODO add test for apifish.segmentation.from_coord_to_surface
